@@ -4,6 +4,8 @@
 from __future__ import annotations
 
 import argparse
+import importlib.abc
+import importlib.machinery
 import importlib.util
 import pathlib
 import sys
@@ -14,20 +16,21 @@ from . import (
     gn,
     grun,
     ty,
+    util,
 )
 
 
 class CLI:
     SD = argparse_subdec.SubDec(name_prefix='__cmd_', fn_dest='core_fn')
 
-    def __init__(self, graph: gn.Graph = None):
+    def __init__(self, graph: ty.Optional[gn.Graph] = None):
         if graph is None:
             graph = gn._global_graph
         self.__graph = graph
         self.__runner = grun.Runner()
         self.__init_parser()
 
-    def run(self, argv: ty.Sequence[str] = None) -> int:
+    def run(self, argv: ty.Optional[ty.Sequence[str]] = None) -> int:
         self.__args = self.__parse_args(argv)
         self.__load_yp()
 
@@ -38,13 +41,13 @@ class CLI:
         self.__args.core_fn(self)
         return 0
 
-    def set_graph(self, graph: gn.Graph):
+    def set_graph(self, graph: gn.Graph) -> None:
         self.__graph = graph
 
-    def set_runner(self, runner: grun.Runner):
+    def set_runner(self, runner: grun.Runner) -> None:
         self.__runner = runner
 
-    def __init_parser(self):
+    def __init_parser(self) -> None:
         self.__parser = argparse.ArgumentParser(description='Run Yape!')
 
         self.__parser.add_argument(
@@ -62,16 +65,19 @@ class CLI:
 
         CLI.SD.create_parsers(self.__subparsers)
 
-    def __parse_args(self, argv: ty.Sequence[str]):
+    def __parse_args(self,
+                     argv: ty.Optional[ty.Sequence[str]],
+                     ) -> argparse.Namespace:
         if argv is None:
             argv = sys.argv[1:]
         argv = list(argv)
         return self.__parser.parse_args(argv)
 
-    def __load_yp(self):
+    def __load_yp(self) -> None:
         name = self.__args.core_yp
         sys.path = ['.'] + sys.path
         try:
+            spec: ty.Optional[importlib.machinery.ModuleSpec]
             spec = importlib.util.find_spec(name)
             if not spec:
                 spec = importlib.util.spec_from_file_location('yp', name)
@@ -80,7 +86,10 @@ class CLI:
                 raise RuntimeError(f'failed to find module or file: {name}')
 
             module = importlib.util.module_from_spec(spec)
-            spec.loader.exec_module(module)
+            loader = ty.cast(importlib.abc.Loader, spec.loader)
+            loader.exec_module(module)
+            if hasattr(module, 'nodegen'):
+                module.nodegen()
         finally:
             if sys.path and sys.path[0] == '.':
                 sys.path = sys.path[1:]
@@ -109,7 +118,7 @@ class CLI:
         Run nodes even if cached results are up to date.
         """
     )
-    def __cmd_run(self):
+    def __cmd_run(self) -> None:
         targets = getattr(self.__args, 'run_targets', None)
         if not targets:
             targets = None
@@ -123,9 +132,45 @@ class CLI:
 
     @SD.cmd(
         description="""
-        List available targets.
+        List available targets. Only explicitly named targets are listed by
+        default. You can use --all to list all targets.
         """
     )
-    def __cmd_list(self):
-        for node in self.__graph.recurse_nodes():
+    @SD.add_argument(
+        '-a', '--all',
+        action='store_true',
+        dest='list_all',
+        help="""
+        List both named and unnamed targets.
+        """
+    )
+    def __cmd_list(self) -> None:
+        pred: ty.Optional[ty.Callable[[gn.Node[ty.Any]], bool]] = None
+        if not self.__args.list_all:
+            def pred(node: gn.Node[ty.Any]) -> bool:
+                return node._has_explicit_name
+        for node in self.__graph.recurse_nodes(pred):
             print(node._fullname())
+
+    @SD.cmd(
+        description="""
+        List dependencies for each node passed as target. All nodes are used as
+        targets by default. You can pass specific targets as positional
+        arguments.
+        """,
+    )
+    @SD.add_argument(
+        'deps_targets',
+        default=None,
+        nargs='*',
+        metavar='TARGET',
+    )
+    def __cmd_deps(self) -> None:
+        targets = getattr(self.__args, 'deps_targets', None)
+        if not targets:
+            targets = None
+        nodes, _ = util.parse_targets(targets, self.__graph)
+        for node in nodes:
+            print(node._fullname())
+            for dep in node._get_dep_nodes():
+                print(f'    {dep._fullname()}')
